/************************************************************************************************
*  Filename: filedeleter.c
*  Author: <kasskloff AT gmail.com>
*  Description: A program to bypass NTFS file permissions and delete files/folders, e.g. those
*  owned by TrustedInstaller.
*  
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*  
*  The above copyright notice and this permission notice shall be included in all
*  copies or substantial portions of the Software.
*  
*  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
*  SOFTWARE.
************************************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <time.h>
#include <windows.h>

#define STRING_EMPTY ""
#define EXIT_NO_EXIT 0
#define EXIT_EXIT 1

#define SHRED_SHRED 1
#define SHRED_NO_SHRED 0

#define ERROR_NO_ERROR 0x00
#define ERROR_NO_PTOKEN 0x01
#define ERROR_NO_PRIVILEGE 0x02
#define ERROR_TOO_LONG 0x03
#define ERROR_NO_OPEN 0x04
#define ERROR_NO_WRITE 0x05
#define ERROR_NO_DELETE 0x06
#define ERROR_NO_ALLOC 0x07

void RunProgramA(int argc, char* argv[]);
void RunProgramW(int argc, wchar_t* argv[]);
void EmptyDirectoryA(const char* directory, int shredFiles);
void EmptyDirectoryW(const wchar_t* directory, int shredFiles);
void ShredFileA(const char* path);
void ShredFileW(const wchar_t* path);
char* FormatSizeA(DWORD bytes);
wchar_t* FormatSizeW(DWORD bytes);
int EnablePrivilege(HANDLE token, LPCTSTR privilege);
char* GetErrorMessageA(DWORD errNo);
wchar_t* GetErrorMessageW(DWORD errNo);
int GetConfirmation();
void DisplayErrorA(DWORD errNo, int errType, const char* file, int doExit);
void DisplayErrorW(DWORD errNo, int errType, const wchar_t* file, int doExit);

int main(int argc, char* argv[])
{
	SetConsoleTitle("File Deleter");
	srand(time(NULL));
	
	typedef BOOL WINAPI (*import_Wow64DisableWow64FsRedirection)(PVOID* OldValue);
	HINSTANCE kernel32;
	import_Wow64DisableWow64FsRedirection Wow64DisableWow64FsRedirection;
	kernel32 = LoadLibrary("kernel32.dll");
	Wow64DisableWow64FsRedirection = (import_Wow64DisableWow64FsRedirection)GetProcAddress(kernel32, "Wow64DisableWow64FsRedirection");
	if (Wow64DisableWow64FsRedirection)
	{
		PVOID oldVal;
		Wow64DisableWow64FsRedirection(&oldVal);
	}
	
	OSVERSIONINFO os;
	memset(&os, 0, sizeof(OSVERSIONINFO));
	os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&os);

	if (os.dwPlatformId == VER_PLATFORM_WIN32_NT)
	{
		HANDLE procTok;
		if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &procTok))
			DisplayErrorA(GetLastError(), ERROR_NO_TOKEN, STRING_EMPTY, EXIT_EXIT);
		
		if (EnablePrivilege(procTok, "SeRestorePrivilege") != 0 || EnablePrivilege(procTok, "SeBackupPrivilege") != 0)
			DisplayErrorA(GetLastError(), ERROR_NO_PRIVILEGE, STRING_EMPTY, EXIT_EXIT);

		CloseHandle(procTok);
	}
	
	DWORD unicodeSupport = (os.dwMajorVersion > 4 ? 1 : 0);
	if (unicodeSupport)
	{
		int argcW;
		wchar_t** argvW;
		argvW = (wchar_t**)CommandLineToArgvW(GetCommandLineW(), &argcW);
		if (argvW != NULL)
			RunProgramW(argcW, argvW);
	}
	RunProgramA(argc, argv);
	
}

void RunProgramA(int argc, char* argv[])
{
	char input[MAX_PATH];
	memset(input, '\0', sizeof(input));
	
	if (argc > 1)
	{
		if (strlen(argv[1]) > MAX_PATH)
			DisplayErrorA(EXIT_FAILURE, ERROR_TOO_LONG, argv[1], EXIT_EXIT);

		strcpy(input, argv[1]);
	}
	else
	{
		printf("\nEnter the path to the file/folder you wish to delete:\n> ");
		fgets(input, sizeof(input), stdin);
		
		if (input[0] == '\0')
			return;
		if (input[strlen(input) - 1] == '\n') 
			input[strlen(input) - 1] = '\0';
	}
	
	DWORD attrib = GetFileAttributes(input);
	if (attrib == INVALID_FILE_ATTRIBUTES)
		DisplayErrorA(GetLastError(), ERROR_NO_OPEN, input, EXIT_EXIT);
	
	int prompt = -1;
	int shred = -1;
	
	for (int i = 1; i < argc; i++)
	{
		char* tempArg = calloc(strlen(argv[i]) + 1, sizeof(char));
		if (tempArg == NULL)
			DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
		
		strcpy(tempArg, argv[i]);
		strlwr(tempArg);

		if (!strcmp(tempArg, "/y"))
			prompt = 0;
		else if (!strcmp(tempArg, "/s") || !strcmp(tempArg, "/shred"))
			shred = 1;
		else if (!strcmp(tempArg, "/ns") || !strcmp(tempArg, "/noshred"));
			shred = 0;
		
		free(tempArg);
	}
	
	if (prompt == -1)
	{
		printf("\nDelete %s [%s] - Are you sure? (Y/N): ", (attrib & FILE_ATTRIBUTE_DIRECTORY ? "directory (and all files/subdirectories)" : "file"), input);
		prompt = GetConfirmation();
		if (!prompt)
			DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
	}
	
	if (shred == -1)
	{
		printf("\nShred %s by overwriting with random data before deleting?", (attrib & FILE_ATTRIBUTE_DIRECTORY ? "files" : "file"));
		printf("\nNote: This will dramatically increase execution time for large files (Y/N): ");
		shred = GetConfirmation();
	}
	
	if (attrib & FILE_ATTRIBUTE_READONLY)
		SetFileAttributes(input, attrib & ~FILE_ATTRIBUTE_READONLY);
	
	if (attrib & FILE_ATTRIBUTE_DIRECTORY)
	{
		EmptyDirectoryA(input, shred);
		if (RemoveDirectory(input))
		{
			printf("\nDirectory [%s] deleted!\n", input);
			DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
		}
		else
			DisplayErrorA(GetLastError(), ERROR_NO_DELETE, input, EXIT_EXIT);
	}
	
	if (shred)
		ShredFileA(input);
	if (DeleteFile(input))
	{
		printf("\nFile [%s] deleted!\n", input);
		DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
	}
	else
		DisplayErrorA(GetLastError(), ERROR_NO_DELETE, input, EXIT_EXIT);
	
}

void RunProgramW(int argc, wchar_t* argv[])
{
	fflush(stdout);
	_setmode(_fileno(stdout), 0x20000);
	wchar_t input[MAX_PATH];
	memset(input, '\0', sizeof(input));
	
	if (argc > 1)
	{
		if (wcslen(argv[1]) > MAX_PATH)
			DisplayErrorW(EXIT_FAILURE, ERROR_TOO_LONG, argv[1], EXIT_EXIT);

		wcscpy(input, argv[1]);
	}
	else
	{
		wprintf(L"\nEnter the path to the file/folder you wish to delete:\n> ");
		fgetws(input, sizeof(input), stdin);
		
		if (input[0] == '\0')
			return;
		if (input[wcslen(input) - 1] == '\n') 
			input[wcslen(input) - 1] = '\0';
	}
	
	DWORD attrib = GetFileAttributesW(input);
	if (attrib == INVALID_FILE_ATTRIBUTES)
	{
		LocalFree((LPWSTR*)argv);
		DisplayErrorW(GetLastError(), ERROR_NO_OPEN, input, EXIT_EXIT);
	}
	
	int prompt = -1;
	int shred = -1;
	
	for (int i = 1; i < argc; i++)
	{
		wchar_t* tempArg = calloc(wcslen(argv[i]) + 1, sizeof(wchar_t));
		if (tempArg == NULL)
			DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
		
		wcscpy(tempArg, argv[i]);
		wcslwr(tempArg);

		if (!wcscmp(tempArg, L"/y"))
			prompt = 0;
		else if (!wcscmp(tempArg, L"/s") || !wcscmp(tempArg, L"/shred"))
			shred = 1;
		else if (!wcscmp(tempArg, L"/ns") || !wcscmp(tempArg, L"/noshred"))
			shred = 0;
		
		free(tempArg);
	}
	
	LocalFree((LPWSTR*)argv);
	
	if (prompt == -1)
	{
		wprintf(L"\nDelete %s [%s] - Are you sure? (Y/N): ", (attrib & FILE_ATTRIBUTE_DIRECTORY ? L"directory (and all files/subdirectories)" : L"file"), input);
		prompt = GetConfirmation();
		if (!prompt)
			DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
	}
	
	if (shred == -1)
	{
		wprintf(L"\nShred %s by overwriting with random data before deleting?", (attrib & FILE_ATTRIBUTE_DIRECTORY ? L"files" : L"file"));
		wprintf(L"\nNote: This will dramatically increase execution time for large files (Y/N): ");
		shred = GetConfirmation();
	}
	
	if (attrib & FILE_ATTRIBUTE_READONLY)
		SetFileAttributesW(input, attrib & ~FILE_ATTRIBUTE_READONLY);
	
	if (attrib & FILE_ATTRIBUTE_DIRECTORY)
	{
		EmptyDirectoryW(input, shred);
		if (RemoveDirectoryW(input))
		{
			wprintf(L"\nDirectory [%s] deleted!\n", input);
			DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
		}
		else
			DisplayErrorW(GetLastError(), ERROR_NO_DELETE, input, EXIT_EXIT);
	}
	
	if (shred)
		ShredFileW(input);
	if (DeleteFileW(input))
	{
		wprintf(L"\nFile [%s] deleted!\n", input);
		DisplayErrorA(EXIT_SUCCESS, ERROR_NO_ERROR, STRING_EMPTY, EXIT_EXIT);
	}
	else
		DisplayErrorW(GetLastError(), ERROR_NO_DELETE, input, EXIT_EXIT);
}

void EmptyDirectoryA(const char* directory, int shredFiles)
{
	WIN32_FIND_DATA file;
	HANDLE handle;
	
	char pattern[MAX_PATH];
	memset(pattern, '\0', sizeof(pattern));
	strcat(pattern, directory);
	if (pattern[strlen(pattern) - 1] != '\\')
		strcat(pattern, "\\*");
	else
		strcat(pattern, "*");
	
	handle = FindFirstFile(pattern, &file);
	if (handle == INVALID_HANDLE_VALUE)
	{
		DisplayErrorA(GetLastError(), ERROR_NO_OPEN, directory, EXIT_NO_EXIT);
		return;
	}
	
	do
	{
		if (strcmp(file.cFileName, ".") && strcmp(file.cFileName, ".."))
		{
			char* fullName = malloc((sizeof(directory) + sizeof(file.cFileName)) * sizeof(char));
			if (fullName == NULL)
				DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
			
			memset(fullName, '\0', sizeof(fullName));
			strcat(fullName, directory);
			if (fullName[strlen(fullName) - 1] != '\\')
				strcat(fullName, "\\");
			strcat(fullName, file.cFileName);
			
			DWORD attrib = GetFileAttributes(fullName);
			if (attrib == INVALID_FILE_ATTRIBUTES)
			{
				DisplayErrorA(GetLastError(), ERROR_NO_OPEN, fullName, EXIT_NO_EXIT);
				free(fullName);
				continue;
			}
			else if (attrib & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (attrib & FILE_ATTRIBUTE_READONLY)
					SetFileAttributes(fullName, attrib & ~FILE_ATTRIBUTE_READONLY);
				EmptyDirectoryA(fullName, shredFiles);
				if (RemoveDirectory(fullName))
					printf("Directory [%s] deleted!\n", fullName);
				else
					DisplayErrorA(GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
			}
			else
			{
				if (attrib & FILE_ATTRIBUTE_READONLY)
					SetFileAttributes(fullName, attrib & ~FILE_ATTRIBUTE_READONLY);
				if (shredFiles)
					ShredFileA(fullName);
				if (DeleteFile(fullName))
					printf("File [%s] deleted!\n", fullName);
				else
					DisplayErrorA(GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
			}
			free(fullName);
		}
	}
	while (FindNextFile(handle, &file));
	
	FindClose(handle);
}

void EmptyDirectoryW(const wchar_t* directory, int shredFiles)
{
	WIN32_FIND_DATAW file;
	HANDLE handle;
	
	wchar_t pattern[MAX_PATH];
	memset(pattern, '\0', sizeof(pattern));
	wcscat(pattern, directory);
	if (pattern[wcslen(pattern) - 1] != '\\')
		wcscat(pattern, L"\\*");
	else
		wcscat(pattern, L"*");
	
	handle = FindFirstFileW(pattern, &file);
	if (handle == INVALID_HANDLE_VALUE)
	{
		DisplayErrorW(GetLastError(), ERROR_NO_OPEN, directory, EXIT_NO_EXIT);
		return;
	}
	
	do
	{
		if (wcscmp(file.cFileName, L".") && wcscmp(file.cFileName, L".."))
		{
			wchar_t* fullName = malloc((sizeof(directory) + sizeof(file.cFileName)) * sizeof(wchar_t));
			if (fullName == NULL)
				DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
			
			memset(fullName, '\0', sizeof(fullName));
			wcscat(fullName, directory);
			if (fullName[wcslen(fullName) - 1] != '\\')
				wcscat(fullName, L"\\");
			wcscat(fullName, file.cFileName);
			
			DWORD attrib = GetFileAttributesW(fullName);
			if (attrib == INVALID_FILE_ATTRIBUTES)
			{
				DisplayErrorW(GetLastError(), ERROR_NO_OPEN, fullName, EXIT_NO_EXIT);
				free(fullName);
				continue;
			}
			else if (attrib & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (attrib & FILE_ATTRIBUTE_READONLY)
					SetFileAttributesW(fullName, attrib & ~FILE_ATTRIBUTE_READONLY);
				EmptyDirectoryW(fullName, shredFiles);
				if (RemoveDirectoryW(fullName))
					wprintf(L"Directory [%s] deleted!\n", fullName);
				else
					DisplayErrorW(GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
			}
			else
			{
				if (attrib & FILE_ATTRIBUTE_READONLY)
					SetFileAttributesW(fullName, attrib & ~FILE_ATTRIBUTE_READONLY);
				if (shredFiles)
					ShredFileW(fullName);
				if (DeleteFileW(fullName))
					wprintf(L"File [%s] deleted!\n", fullName);
				else
					DisplayErrorW(GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
			}
			free(fullName);
		}
	}
	while (FindNextFileW(handle, &file));
	
	FindClose(handle);
}

void ShredFileA(const char* path)
{
	HANDLE file = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (file == INVALID_HANDLE_VALUE)
	{
		DisplayErrorA(GetLastError(), ERROR_NO_OPEN, path, EXIT_NO_EXIT);
		return;
	}
	unsigned char randomData[4096];
	DWORD fileSize = GetFileSize(file, NULL);
	int failed = 0;
	
	if (!fileSize)
	{
		printf("\nSkipping shredding of file [%s] as it is empty\n", path);
		CloseHandle(file);
		return;
	}
	
	char* sizeStr = FormatSizeA(fileSize);
	printf("\nShredding file [%s] (%s)\n", path, sizeStr);
	free(sizeStr);
	for (int i = 0; i < 3; i++)
	{
		if (failed)
			break;
		
		int final = (i == 2);
		printf("Writing %s [Pass: %d/%d]\n", (final ? "zeroes" : "random bytes"), i + 1, 3);
		if (final)
		{
			for (size_t i = 0; i < 4096; i++)
				randomData[i] = 0;
		}
		
		DWORD totalWritten = 0;
		DWORD written = 0;
		SetFilePointer(file, 0, NULL, FILE_BEGIN);
		do
		{
			if (!final)
			{
				for (size_t i = 0; i < 4096; i++)
					randomData[i] = rand() % 256;
			}
			
			DWORD toWrite = ((fileSize - totalWritten) < 4096 ? (fileSize - totalWritten) : 4096);
			if (!WriteFile(file, randomData, toWrite, &written, NULL))
			{
				DisplayErrorA(GetLastError(), ERROR_NO_WRITE, path, EXIT_NO_EXIT);
				failed = 1;
				CloseHandle(file);
				break;
			}
			totalWritten += written;
			
		} while (totalWritten < fileSize);
	}
	CloseHandle(file);
	
	return;
}

void ShredFileW(const wchar_t* path)
{
	HANDLE file = CreateFileW(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (file == INVALID_HANDLE_VALUE)
	{
		DisplayErrorW(GetLastError(), ERROR_NO_OPEN, path, EXIT_NO_EXIT);
		return;
	}
	unsigned char randomData[4096];
	DWORD fileSize = GetFileSize(file, NULL);
	int failed = 0;
	
	if (!fileSize)
	{
		wprintf(L"\nSkipping shredding of file [%s] as it is empty\n", path);
		CloseHandle(file);
		return;
	}
	
	wchar_t* sizeStr = FormatSizeW(fileSize);
	wprintf(L"\nShredding file [%s] (%s)\n", path, sizeStr);
	free(sizeStr);
	for (int i = 0; i < 3; i++)
	{
		if (failed)
			break;
		
		int final = (i == 2);
		wprintf(L"Writing %s [Pass: %d/%d]\n", (final ? L"zeroes" : L"random bytes"), i + 1, 3);
		if (final)
		{
			for (size_t i = 0; i < 4096; i++)
				randomData[i] = 0;
		}
		
		DWORD totalWritten = 0;
		DWORD written = 0;
		SetFilePointer(file, 0, NULL, FILE_BEGIN);
		do
		{
			if (!final)
			{
				for (size_t i = 0; i < 4096; i++)
					randomData[i] = rand() % 256;
			}
			
			DWORD toWrite = ((fileSize - totalWritten) < 4096 ? (fileSize - totalWritten) : 4096);
			if (!WriteFile(file, randomData, toWrite, &written, NULL))
			{
				DisplayErrorW(GetLastError(), ERROR_NO_WRITE, path, EXIT_NO_EXIT);
				failed = 1;
				CloseHandle(file);
				break;
			}
			totalWritten += written;
			
		} while (totalWritten < fileSize);
	}
	CloseHandle(file);
	
	return;
}

char* FormatSizeA(DWORD bytes)
{
	char* formatted = malloc(64 * sizeof(char));
	if (formatted == NULL)
		DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
	
	if (bytes < 1024)
		sprintf(formatted, "%lu bytes", bytes);
	else if (bytes < 1048576)
		sprintf(formatted, "%lu KB", bytes / 1024);
	else if (bytes < 1073741824)
		sprintf(formatted, "%lu MB", bytes / 1048576);
	else
		sprintf(formatted, "%lu GB", bytes / 1073741824);
	return formatted;	
}

wchar_t* FormatSizeW(DWORD bytes)
{
	wchar_t* formatted = malloc(64 * sizeof(wchar_t));
	if (formatted == NULL)
		DisplayErrorA(EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
	
	if (bytes < 1024)
		swprintf(formatted, L"%lu bytes", bytes);
	else if (bytes < 1048576)
		swprintf(formatted, L"%lu KB", bytes / 1024);
	else if (bytes < 1073741824)
		swprintf(formatted, L"%lu MB", bytes / 1048576);
	else
		swprintf(formatted, L"%lu GB", bytes / 1073741824);
	return formatted;	
}

int EnablePrivilege(HANDLE token, const char* privilege)
{
	TOKEN_PRIVILEGES privilegeToken;
	LUID luid;
	LookupPrivilegeValue(NULL, (LPCTSTR)privilege, &luid);
	privilegeToken.PrivilegeCount = 1;
	privilegeToken.Privileges[0].Luid = luid;
	privilegeToken.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	AdjustTokenPrivileges(token, FALSE, &privilegeToken, sizeof(TOKEN_PRIVILEGES), NULL, NULL);
	
	return GetLastError();
}

char* GetErrorMessageA(DWORD errNo)
{
	LPVOID buffer;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errNo, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&buffer, 0, NULL);
	char* message = (char*)buffer;
	if (message[strlen(message) - 2] == '\r')
		message[strlen(message) - 2] = '\0';
	
	return message;
}

wchar_t* GetErrorMessageW(DWORD errNo)
{
	
	LPVOID buffer;
	FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errNo, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPWSTR)&buffer, 0, NULL);
	wchar_t* message = (wchar_t*)buffer;
	if (message[wcslen(message) - 2] == '\r')
		message[wcslen(message) - 2] = '\0';
	
	return message;
}

int GetConfirmation()
{
	char confirm[3];
	fgets(confirm, sizeof(confirm), stdin);
	return (confirm[0] == 'y' || confirm[0] == 'Y');
}

void DisplayErrorA(DWORD errNo, int errType, const char* file, int doExit)
{
	if (errNo != ERROR_NO_ERROR)
	{
		char* errMsg;
		if (errType != ERROR_NO_ALLOC)
			errMsg = GetErrorMessageA(errNo);
		switch (errType)
		{
			case ERROR_NO_PTOKEN:
				printf("\nCannot get process token ((%u) %s)\n", (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_PRIVILEGE:
				printf("\nCannot adjust privileges ((%u) %s)\n", (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_TOO_LONG:
				printf("\nPath [%s] exceeds maximum path length of %u characters\n", file, MAX_PATH);
			break;

			case ERROR_NO_OPEN:
				printf("\nCannot open file/folder [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_WRITE:
				printf("\nCannot write to file [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;

			case ERROR_NO_DELETE:
				printf("\nCould not delete file/folder [%s] ((%u) %s)\nMake sure the file is not in use and try again\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_ALLOC:
				printf("\nFatal Error: Cannot allocate memory\n");
			break;
			
			default:
				printf("\nAn unknown error occurred\n");
		}
		if (errType != ERROR_NO_ALLOC)
			LocalFree(errMsg);
	}
	
	if (doExit == EXIT_EXIT)
		exit(errNo);

}

void DisplayErrorW(DWORD errNo, int errType, const wchar_t* file, int doExit)
{
	if (errNo != ERROR_NO_ERROR)
	{
		wchar_t* errMsg;
		if (errType != ERROR_NO_ALLOC)
			errMsg = GetErrorMessageW(errNo);
		switch (errType)
		{
			case ERROR_NO_PTOKEN:
				wprintf(L"\nCannot get process token ((%d) %s)\n", errNo, errMsg);
			break;
			
			case ERROR_NO_PRIVILEGE:
				wprintf(L"\nCannot adjust privileges ((%d) %s)\n", errNo, errMsg);
			break;
			
			case ERROR_TOO_LONG:
				wprintf(L"\nPath [%s] exceeds maximum path length of %u characters\n", file, MAX_PATH);
			break;

			case ERROR_NO_OPEN:
				wprintf(L"\nCannot open file/folder [%s] ((%d) %s)\n", file, errNo, errMsg);
			break;

			case ERROR_NO_DELETE:
				wprintf(L"\nCould not delete file/folder [%s] ((%d) %s)\nMake sure the file is not in use and try again\n", file, errNo, errMsg);
			break;
			
			case ERROR_NO_ALLOC:
				wprintf(L"\nFatal Error: Cannot allocate memory\n");
			break;
			
			default:
				wprintf(L"\nAn unknown error occurred\n");
		}
		if (errType != ERROR_NO_ALLOC)
			LocalFree(errMsg);
	}
	
	if (doExit == EXIT_EXIT)
		exit(errNo);

}
